import mbusif
import mbustypes
import time

# TODO: Add Timeout to RPCs

# Notes: Currently only RPCs are supported
# FIXME how/when to send reliably rpc's?
# TODO add RPC-STATUS to rpc returns

class Guidelines( mbusif.MbusIF ):
    """Extends the Mbus transport layer with the guidelinefunctionality."""
    def __init__( self, addr ):
	mbusif.MbusIF.__init__( self, addr )
	self.rpcReturnCallbacks = {} # RPC-Return-IDs -> callBack-Functions
	self.registeredRPCs = {}     # RPC-Command -> Callback
	self.multicastRPCReturns = {}# MulticastRPCReturn buffer
	self.rpcTimeout = 2.0	     # 2 seconds max.
    def addRPCCallback( self, cmdname, callback, data = None ):
	"""Register callbacks for rpc-commands. First argument specifies the name of
	the rpc-command. The second is the callback that is to be invoked and the
	third one is optional and can be used to pass data to callback function.
	The callback function has to take three arguments. The senders address, the
	arguments of the command and a the optional data.
	This callback function must return a RPCReturn object. To signal the caller
	a failure just raise a RPCException containing the errormessage. If you
	want to silently ignore the rpccall (send no return message) then simply
	return None."""
	self.registeredRPCs[ cmdname ] = ( callback, data )
	self.addCallback( cmdname, self.incomingRPC )
    def incomingRPC( self, mmsg, data = None ):
	"""Internal function."""
	returnMsg = mbustypes.MMessage( \
		mbustypes.maddress( mmsg.header.srcAdr ) )
	returnMsg.payload.append( mbustypes.MCommand( \
		"%s.return"%( mmsg.payload[0].name ) ) )
	# FIXME honor RPC-TYPE
	rpcType = None
	rpcID = None
	if len( mmsg.payload[0].args ) == 2:
	    for arg in mmsg.payload[0].args[0]:
		if len( arg ) == 2:
		    if arg[0] == "RPC-TYPE":
			rpcType = arg[1]
		    if arg[0] == "ID":
			rpcID = arg[1]
	if rpcID == None:
	    if self.errorFunction:
		self.errorFunction( mbustypes.MbusError( \
			"invalid rpc-call: missing id", mmsg ) )
	    return 1
	if not self.registeredRPCs.has_key( mmsg.payload[0].name ):
	    returnMsg.payload[-1].args = \
		    [ [[ "ID", rpcID ], [ "RPC-STATUS", "UNKNOWN" ]], [ ] ]
	    self.send( returnMsg )
	    if self.errorFunction:
		self.errorFunction( mbustypes.MbusError( \
			"received unknown rpc-command", mmsg ) )
		return
	try:
	    result = self.registeredRPCs[ mmsg.payload[0].name ][0]\
		    ( mmsg.header.srcAdr, mmsg.payload[0].args[1], \
		      self.registeredRPCs[ mmsg.payload[0].name ][1] )
	except RPCException, e:
	    result = RPCReturn( [], "a", e.errormsg )
	    result.appStatus = e.status
	    result.rpcStatus = mbustypes.MSymbol( "FAILED" )
	if result == None:
	    return 1 # doesn't want to send return
	#returnMsg.payload[-1].args = [ mmsg.payload[0].args[0], [ result ] ]
	returnMsg.payload[-1].args = \
		[ [[ "ID", rpcID ], [ "RPC-STATUS", "OK" ]], result.asMList() ]
	#print "SENDRPCRETURN:",returnMsg.asString()
	self.send( returnMsg )
	return 1
    def sendRPC( self, addr, cmdname, cmdargs, callback, type = "UNICAST", data = None ):
	"""Send rpc-command to entity/entities specified by first arg (an maddress).
	The second argument specifies the commandname and the third it's parameters.
	The fourth argument is a callback-function which gets called on incoming
	results to this rpc-call. It has to take 1 argument, which is a list of pairs.
	In case of a unicast or anycast rpc the list will always only have one
	element, in case of a multicast rpc the list may be longer. Each element of
	this list is a triple of ( address, status, result ).
	The last parmater of the sendRPC method is optional and may be used
	to specify the RPC-TYPE (see guidelines). It defaults to UNICAST."""
	mmsg = mbustypes.MMessage( addr )
	mmsg.payload.append( mbustypes.MCommand( cmdname ) )
	rpcID = "%d"%( mmsg.header.seqNum )
	rpcargs = [ [ "ID", rpcID ], [ "RPC-TYPE", type ] ]
	if type == "UNICAST": mmsg.header.reliable = "R" # RECOMMENDED in mbusguides
	self.addTimer( self.rpcTimeout, self.timeoutRPCReturn, rpcID )
	mmsg.payload[-1].args = [ rpcargs, cmdargs ]
	self.rpcReturnCallbacks[ rpcID ] = ( callback, type, data )
	self.addCallback( "%s.return"%(cmdname), self.incomingRPCReturn )
	#print "RPCSEND:",mmsg.asString()
	self.send( mmsg )
    def incomingRPCReturn( self, mmsg, data = None ):
	"""Internal function."""
	try:
	    rpcid = ""
	    for arg in mmsg.payload[0].args[0]:	# get RPC-ID
		if arg[0] == "ID":
		    rpcid = arg[1]
		    break
	    else: # id not found
		if self.errorFunction:
		    self.errorFunction( mbustypes.MbusError( "missing id in rpc return", mmsg ) )
	    if self.rpcReturnCallbacks.has_key( rpcid ):
		if self.rpcReturnCallbacks[ rpcid ][1] == "UNICAST":
		    self.rpcReturnCallbacks[ rpcid ][0]\
			    ( [( mmsg.header.srcAdr, mmsg.payload[0].args[1][0],\
			     mmsg.payload[0].args[1][1] )], self.rpcReturnCallbacks[rpcid][2] )
		    del self.rpcReturnCallbacks[ rpcid ]
		elif self.rpcReturnCallbacks[ rpcid ][1] == "MULTICAST":
		    if self.multicastRPCReturns.has_key( rpcid ):
			self.multicastRPCReturns[ rpcid ].append\
			    ( ( mmsg.header.srcAdr, mmsg.payload[0].args[1][0],\
			     mmsg.payload[0].args[1][1] ) )
		    else:
			self.multicastRPCReturns[ rpcid ] = \
			    [ ( mmsg.header.srcAdr, mmsg.payload[0].args[1][0],\
			     mmsg.payload[0].args[1][1] ) ]
		elif self.rpcReturnCallbacks[ rpcid ][1] == "ANYCAST":
		    pass # TODO someday..............
	#    else:
	#	if self.errorFunction:
	#	    self.errorFunction( mbustypes.MbusError( "missing callback for rpc return", mmsg ) )
	except IndexError:
	    if self.errorFunction:
		self.errorFunction( mbustypes.MbusError( "invalid rpc return", mmsg ) )
	return 1
    def timeoutRPCReturn( self, rpcID ):
	if self.rpcReturnCallbacks.has_key( rpcID ):
	    if self.rpcReturnCallbacks[ rpcID ][1] == "UNICAST":
		if self.errorFunction:
		    self.errorFunction( mbustypes.MbusError( "timeout while waiting on rpc return. RPC-ID: %s"%( rpcID ) ) )
	    elif self.rpcReturnCallbacks[ rpcID ][1] == "MULTICAST":
		if self.multicastRPCReturns.has_key( rpcID ):
		    self.rpcReturnCallbacks[ rpcID ][0]\
			    ( self.multicastRPCReturns[ rpcID ], \
			    self.rpcReturnCallbacks[rpcID][2] )
		    del self.multicastRPCReturns[ rpcID ]
		else:
		    if self.errorFunction:
			self.errorFunction( mbustypes.MbusError( \
			"timeout while waiting on rpc return. RPC-ID: %s"%( rpcID ) ) )
	    elif self.rpcReturnCallbacks[ rpcID ][1] == "ANYCAST":
		if self.errorFunction:
		    self.errorFunction( mbustypes.MbusError( "timeout while waiting on rpc return. RPC-ID: %s"%( rpcID ) ) )
	    del self.rpcReturnCallbacks[ rpcID ]
    # Minimal register functionality for use with librepcom
    def register( self, addr, prefix, callback, data = None ):
	"""Register prefix. After registration callback gets called with the
	three paramters: status (MSymbol OK/FAILED), app_specific_status
	(MSymbol OK, DENIED, NO_SUCH_COMMAND, MAX_CLIENTS etc), description
	(human readable string)."""
	print mbustypes.maddress( self.address ) 
	self.sendRPC( addr, "mbus.register", [ prefix, mbustypes.maddress( self.address ) ], self.register_return, "UNICAST", [ callback, data ] )
    def register_return( self, args, data ):
	addr, status, result = args[0]
	callback, data = data
	callback( status[0], status[1], status[2], data )
    def deregister( self, addr, prefix, callback, data = None ):
	"""Unregister prefix. After registration callback gets called with the
	three paramters: status (MSymbol OK/FAILED), app_specific_status (MSymbol),
	description (human readable string)."""
	self.sendRPC( addr, "mbus.deregister", [ prefix, mbustypes.maddress( self.address ) ], self.deregister_return, "UNICAST", [ callback, data ] )
    def deregister_return( self, args, data ):
	addr, status, result = args[0]
	callback, data = data
	callback( status[0], status[1], status[2], data )

class RPCException:
    """RPCException may be raised from with the RPC-Callback (server-side)."""
    def __init__( self, status, errormsg = "failed" ):
	"""The first parameter (status) must be a single word describing the type of failure, The second
	parameter (errormsg) should be a human readable description of the
	problem."""
	self.errormsg = errormsg
	self.status = mbustypes.MSymbol( status )
class RPCReturn:
    """A RPCReturn object has to be returned by the RPC-Callback (server-side)."""
    def __init__( self, result, status = "OK", descr = "success" ):
	"""The first parameter (result) has to be a MList containing the result of
	the function call. The second parameter (status) is optional and must be a
	single word describing the return status of the called function. The third
	parameter (descr) is also optional and should be a string describing the
	status in a humanreadable form."""
	self.appStatus = mbustypes.MSymbol( status )
	self.descr = descr
	self.result = result
	self.rpcStatus = mbustypes.MSymbol( "OK" )
    def asMList( self ):
	return [ [ self.rpcStatus, self.appStatus, self.descr ], self.result ]
