from string import split, strip
from random import random
import sys
import os, socket, time, base64, hmac
import notifier, mbuscfg, mbustypes

# transport draft constants (in python every timevalue is measured in seconds)
c_hello_factor     = .2
c_hello_min        = 1
c_hello_dither_min = 0.9
c_hello_dither_max = 1.1
c_hello_dead       = 5
T_r                = 0.1
N_r                = 3
T_c                = 0.07
T_k                = (N_r * (N_r + 1) / 2) * T_r

def nil_fct():
    pass # dummy function

class MbusIF:
    """Mbus Interface -- Handles low level mbus transport functionality"""
    seqNum = 0
    instanceCounter = 0
    def __init__( self, addr ):
	"""The adress given as only argument may either be specified as string or hash"""
	"""This function is called on unknown message. It gets one parameter:
	the MMessage."""
	self.unknownMessageFunction = None
	"""This function is called when new entities enter the MBus.
	It gets called with one parameter: the address of the new entity"""
	self.newEntityFunction      = None
	"""This function is called whenever an entity gets lost (has left the
	mbus). It gets one parameter: the address of the lost entity"""
	self.lostEntityFunction     = None
	"""This function is called whenever mbus errors occur. It gets the
	MbusError as only parameter."""
	self.errorFunction          = None
	self.callbacks = {}
	self.hello_p = 0	# timestamp of last mbus.hello transmit
	self.knownEntities = {}	# each key is a full entity address, each value
				# is a pair containing the timestamp of its
				# last message and its IP-Addr
	self.reliableMsgs = {}	# contains reliable msgs still need to be acked
	self.ackListBak = {}	# timestamps of sended acks, keys = "%d|%s"%(seqnum,addr)
	# append Notifer functions to MbusIF-instance
	# (MbusIF no longer inherits Notifier)
	for i in ["addSocket","removeSocket","addTimer","removeTimer","step","loop"]:
	    if notifier.__dict__.has_key( i ):
		self.__dict__[i] = notifier.__dict__[i]
	    else:
		print "error in notifier"
		sys.exit(1)
	self.reconfigure( addr )
    def reconfigure( self, newAddr, cfg = mbuscfg.MbusCfg() ):
	"""reconfigure changes the mbus address and the mbus parameters
	(e.g. multicast address, port, hashkey etc) of the entity at
	runtime"""
	if self.__dict__.has_key( socket ): self.shutdown();
	# read config
	self.cfg = mbuscfg.MbusCfg()
	if split( self.cfg.hashkey[1:-1], "," )[0] == "HMAC-MD5-96":
	    mbustypes.hashalgo = mbustypes.md5
	mbustypes.hashkey = split( self.cfg.hashkey[1:-1], "," )[1]
	# address parsing
	self.address = {}
	if type( newAddr ) == type( "string" ):
	    for i in split( newAddr ):
		key, value = split( i, ":" )
		self.address[key] = value
	elif type( newAddr ) == type( {} ):
	    self.address = newAddr
	else:
	    raise "addres must be string or hash"
	if not self.address.has_key( "id" ):
	    self.address[ "id" ] = "%d-%d@%s"%( os.getpid(),
		MbusIF.instanceCounter,
		socket.gethostbyname( socket.gethostname() ) )
	    MbusIF.instanceCounter += 1
	# open mcast socket and add to listener
	self.mcastsocket = openmcastsock( self.cfg.addr, self.cfg.port )
	if self.cfg.scope == "LINKLOCAL":
	    #self.mcastsocket.setsockopt( socket.SOL_IP, socket.IP_TTL, 1 )
	    self.mcastsocket.setsockopt( socket.SOL_IP, socket.IP_MULTICAST_TTL, 1 )
	elif self.cfg.scope == "HOSTLOCAL":
	    #self.mcastsocket.setsockopt( socket.SOL_IP, socket.IP_TTL, 0 )
	    # FIXME does IP_MULTICAST_TTL do anything???
	    self.mcastsocket.setsockopt( socket.SOL_IP, socket.IP_MULTICAST_TTL, 0 )
	else:
	    raise mbustypes.MbusError( "unknown scope %s"%(self.cfg.scope) )
	self.addSocket( self.mcastsocket, self.recvIncomingMessage )
	# open sender and unicast receive socket
	self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(('', 0))
	self.addSocket( self.socket, self.recvIncomingMessage )
	self.sendHello()
	self.addTimer( 1, self.expireKnownEntities )
    def recvIncomingMessage( self, sock, data = None):
	"""Internal Function. New messages are received and parsed here."""
	# The data parameter is only for pleasing gtk-mainloop!
	msg, sender = sock.recvfrom( 256000 )
	while ord( msg[-1] ) == 0:
	    msg = msg[:-1]
	saddr,daddr =  extractAddress( msg )
	saddr =  mbustypes.maddress( saddr )
	daddr =  mbustypes.maddress( daddr )
	if saddr == self.address: return # ignore our own broadcast messages
	# add entity / update its timestamp
	srcAdr = ""
	for key in saddr.keys():
	    srcAdr += "%s:%s "%( key, saddr[key] )
	if not self.knownEntities.has_key( srcAdr ):
	    self.knownEntities[ srcAdr ] = ( time.time(), sender )
	    if self.newEntityFunction:
		self.newEntityFunction( saddr )
	else: self.knownEntities[ srcAdr ] = ( time.time(), sender )
	# is messaged addressed to us?
	if isAddressedTo( daddr, self.address ):
	    try:
		mmsg = mbustypes.MMessage( msg )
	    except mbustypes.MbusError, e:
		if self.errorFunction:
		    self.errorFunction( e )
		return
	    except( TypeError, ValueError ):
		if self.errorFunction:
		    self.errorFunction( mbustypes.MbusError( "Parser error", msg ) )
		return
	    self.incomingMessage( sender, mmsg )
    def incomingMessage( self, sender, mmsg ):
	"""Internal Function. Received messages are handled here."""
	# check for acknowledgements
	for seqNum in mmsg.header.ackList:
	    if self.reliableMsgs.has_key( seqNum ):
		del self.reliableMsgs[ seqNum ]
	ackID = "%d|%s"%(mmsg.header.seqNum,mbustypes.maddress(mmsg.header.srcAdr)) 
	# check wether we have to resend an ack
	if self.ackListBak.has_key( ackID ):
	    m = mbustypes.MMessage( mmsg.header.srcAdr )
	    m.header.ackList.append( mmsg.header.seqNum )
	    self.ackListBak[ ackID ] = time.time()
	    self.send( m )
	    return # already processed this message
	# expire ackList Backup
	now = time.time()
	for k in self.ackListBak.keys():
	    if now - self.ackListBak[k] > T_k:
		del self.ackListBak[k]
	# check for reliability
	if mmsg.header.reliable == "R":
	    m = mbustypes.MMessage( mmsg.header.srcAdr )
	    m.header.ackList.append( mmsg.header.seqNum )
	    self.ackListBak[ackID] = time.time()
	    self.send( m )
	# parse commands
	for mcmd in mmsg.payload:
	    # first check for mbus control messages
	    isMbusCtrlMsg = 1
	    if   mcmd.name == "mbus.hello":
		pass # we've already updated senders timestamp
	    elif mcmd.name == "mbus.ping":
		self.sendHello( 1 )
	    elif mcmd.name == "mbus.bye":
		if self.knownEntities.has_key( mbustypes.maddress( mmsg.header.srcAdr ) ):
		    del self.knownEntities[ mbustypes.maddress( mmsg.header.srcAdr ) ]
		    if self.lostEntityFunction:
			self.lostEntityFunction( mmsg.header.srcAdr )
	    elif mcmd.name == "mbus.quit":
		pass # not honored at the moment
	    elif mcmd.name == "mbus.waiting":
		pass # TODO not yet implemented
	    elif mcmd.name == "mbus.go":
		pass # TODO not yet implemented
	    else:
		isMbusCtrlMsg = 0
	    # send command messages to user
	    for cmd in mmsg.payload:
		m = mmsg
		m.payload = [cmd]
		if self.callbacks.has_key( cmd.name ):
		    if not self.callbacks[ cmd.name ][0]( \
				    m, self.callbacks[cmd.name][1] ):
			del self.callbacks[ cmd.name ]
		elif not isMbusCtrlMsg:
		    if self.unknownMessageFunction:
			self.unknownMessageFunction( m )
    def sendHello( self, force = 0 ):
	"""Send a mbus.hello broadcast message. The optional parameter can be used
	to force the transmission. If ommited hellos are only send in scaling
	interval."""
	hello_d = max( c_hello_min, len(self.knownEntities) * c_hello_factor )
	hello_e = hello_d * ( c_hello_dither_max - random()/5 )
	msg = mbustypes.MMessage( "" )
	msg.payload.append( mbustypes.MCommand( "mbus.hello" ) )
	unicast = self.cfg.unicast; self.cfg.unicast = 0 # force multicast on hellos
	self.send( msg )
	self.cfg.unicast = unicast
	self.hello_p = time.time()
	#print "DEBUG: Next Hello in %f (%f) secs (ents: %d)"%(hello_e,hello_d,len(self.knownEntities)+1)
	if force:
	    self.removeTimer( self.sendHello )
	self.addTimer( hello_e, self.sendHello )
	return 0 # This function is a onetimeshot
    def expireKnownEntities( self, data ):
	"""Internal function that expires entities which disappeared from the mbus."""
	now = time.time()
	hello_d = max(c_hello_min, (len(self.knownEntities)+1) * c_hello_factor)
	isDead = c_hello_dead * ( hello_d * c_hello_dither_max )
	for adrStr in self.knownEntities.keys():
	    if self.knownEntities[adrStr][0] + isDead < now:
		del self.knownEntities[adrStr]
		if self.lostEntityFunction:
		    addr = {}
		    for i in split( adrStr ):
			key, value = split( i, ":" )
			addr[key] = value
		    self.lostEntityFunction( addr )
	self.addTimer( isDead, self.expireKnownEntities )
	return 0 # This function is a onetimeshot
    def addCallback( self, mcommand, function, data = None ):
	"""If an MCommand with the name of argument 1 is received the function
	given as second argument is called with the full MMessage as
	Argument. The function has to return true if it wants to be called
	again if another command by the same name is received later."""
	self.callbacks[ mcommand ] = ( function, data )
    def isUniqAddress( self, addr ):
	"""If given address is (at the moment) a uniq mbus address the IP-Address
	(Pair of IP-Addr and Port) are return, else None is returned."""
	cnt = 0
	res = None
	for a in self.knownEntities.keys():
	    if isAddressedTo( addr, mbustypes.maddress( a ) ):
		cnt += 1
		res = self.knownEntities[a][1]
	if cnt == 1:
	    return res
	else:
	    return None
    def send( self, mmsg_or_addr, cmdname = None, cmdargs = None ):
	"""Transmit MMessage over mbus. This function may be called two ways.
	You can give the entire MMessage as only parameter, or specify the message
	content by giving an mbusaddress, a command name and the command arguments
	as one list."""
	if isinstance( mmsg_or_addr, mbustypes.MMessage ):
	    mmsg = mmsg_or_addr
	else:
	    mmsg = mbustypes.MMessage( mmsg_or_addr )
	    mmsg.payload.append( mbustypes.MCommand( cmdname ) )
	    mmsg.payload[-1].args = cmdargs
	if not mmsg.header.srcAdr: mmsg.header.srcAdr = self.address
	uniqIPaddr = "unchecked"
	addr = ( self.cfg.addr, self.cfg.port ) # sending socket
	if self.cfg.unicast:
	    uniqIPaddr = self.isUniqAddress( mmsg.header.destAdr )
	    if uniqIPaddr: # send msg via unicast udp
		addr = uniqIPaddr
	    #print "sending to",addr
	# handle reliable messages
	if mmsg.header.reliable == "R":
	    if uniqIPaddr == "unchecked":
		uniqIPaddr = self.isUniqAddress( mmsg.header.destAdr )
	    if not uniqIPaddr:
		raise mbustypes.MbusError( \
			"couldn't send reliable, target is not uniq", mmsg )
	    if mmsg.header.__dict__.has_key( "retransmissionCounter" ):
		if mmsg.header.retransmissionCounter > N_r:
		    if self.errorFunction:
			self.errorFunction( mbustypes.MbusError( \
				"Message acknowledgement failed", mmsg ) )
			del self.reliableMsgs[ mmsg.header.seqNum ] ; return
		else: mmsg.header.retransmissionCounter += 1
	    else: mmsg.header.retransmissionCounter = 1
	    self.reliableMsgs[ mmsg.header.seqNum ] = mmsg
	    self.addTimer( T_r * mmsg.header.retransmissionCounter, \
		    self.resendReliable, mmsg.header.seqNum )
	buf = mmsg.asString()
	size = 0
	try:
	    while size < len( buf ):
		size += self.socket.sendto( buf[size:], addr )
	except socket.error:
	    raise notifier.DeadTimerException()
    def sendReliable( self, mmsg_or_addr, cmdname = None, cmdargs = None ):
	"""Transmit MMessage reliable over mbus. Same parameters as the basic send
	function."""
	if isinstance( mmsg_or_addr, mbustypes.MMessage ):
	    mmsg = mmsg_or_addr
	else:
	    mmsg = mbustypes.MMessage( mmsg_or_addr )
	    mmsg.payload.append( mbustypes.MCommand( cmdname ) )
	    mmsg.payload[-1].args = cmdargs
	mmsg.header.reliable = "R"
	self.send( mmsg )
    def resendReliable( self, seqNum ):
	"""Internal function to resend a reliable message."""
	try:
	    if self.reliableMsgs.has_key( seqNum ):
		self.send( self.reliableMsgs[ seqNum ] )
	except mbustypes.MbusError,e:
	    if self.errorFunction: self.errorFunction( e )
	return 0
    def shutdown( self ):
	self.send( "", "mbus.bye", [] )
	self.removeSocket( self.socket )
	self.socket.close()
	self.removeSocket( self.mcastsocket )
	self.mcastsocket.close()
    def __del__( self ):
	self.shutdown()

    # DEBUG
    def sendraw( self, targetAddress, buf ):
	"""For debugging only! Transmits unchecked raw body"""
	addr = ( self.cfg.addr, self.cfg.port ) # sending socket
	if self.cfg.unicast:
	    uniqIPaddr = self.isUniqAddress( mbustypes.maddress( targetAddress ) )
	    if uniqIPaddr: # send msg via unicast udp
		addr = uniqIPaddr
	buf = mbustypes.MHeader( mbustypes.maddress( targetAddress ) ).asString() + "\n" + buf
	digest = base64.encodestring( hmac.hmac( buf, \
		base64.decodestring( mbustypes.hashkey ), \
		mbustypes.hashalgo ) )[:16]
	buf = "%s\n%s"%( digest, buf )
	size = 0
	while size < len( buf ):
	    s =  self.socket.sendto( buf[size:], addr )
	    print buf[size:size+s], # debug output
	    size += s

# misc util functions

def isAddressedTo( subset, fulladdr ):
    """If the first mbus address matches the second one (also partly) true is
    returned. False otherwise. Both addresses have to be given as hashes!"""
    #if subset == {}:
    #   return 1
    for key in subset.keys():
	if not( fulladdr.has_key( key ) \
		and fulladdr[key] == subset[key] ):
	    return 0
    return 1

def extractAddress( mmsg ):
    """Extract source and destination address message header. This is also done
    by mbustypes.MHeader, but using this function we can check addreses
    before checking the hmac, which is quite time consuming.
    Returns a pair containing source and destination addres normalized string
    form."""
    header = split( mmsg, "\n" )[1]
    version, seqNum, timestamp, reliable, addrs = split( header, None, 4 )
    srcAdr, destAdr, ackList = split( addrs, ")", 2 )
    srcAdrDict = mbustypes.maddress( strip( srcAdr )[1:] )
    destAdrDict = mbustypes.maddress( strip( destAdr )[1:] )
    srcAdr = ""; destAdr = ""
    for key in srcAdrDict.keys():
	srcAdr += "%s:%s "%( key, srcAdrDict[key] )
    for key in destAdrDict.keys():
	destAdr += "%s:%s "%( key, destAdrDict[key] )
    return srcAdr, destAdr

def openmcastsock(group, port):
    """Open a UDP socket, bind it to a port and select a multicast group."""
    # Import modules used only here
    import string
    import struct
    #
    # Create a socket
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    #
    # Allow multiple copies of this program on one machine
    # (not strictly needed)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    #
    # Bind it to the port
    s.bind(('', port))
    #
    # Look up multicast group address in name server
    # (doesn't hurt if it is already in ddd.ddd.ddd.ddd format)
    group = socket.gethostbyname(group)
    #
    # Construct binary group address
    bytes = map(int, string.split(group, "."))
    grpaddr = 0
    for byte in bytes: grpaddr = (grpaddr << 8) | byte
    #
    # Construct struct mreq from grpaddr and ifaddr
    ifaddr = socket.INADDR_ANY
    mreq = struct.pack('ll', socket.htonl(grpaddr), socket.htonl(ifaddr))
    #
    # Add group membership
    s.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
    #
    return s
