#!/usr/bin/env python

from mbus import mbusif
from mbus import mbustypes

def sendReliableTest( data ):
    m = mbustypes.MMessage( "lang:python" ) # target address
    # payload consists of one or more commands
    m.payload.append( mbustypes.MCommand( "mbus.reliable" ) )
    m.payload[-1].args = [ "just", mbustypes.MSymbol( "a" ), "test", [ 1,2,3 ] ]
    print "sending reliable message"
    try:
	mbus.sendReliable( m )
    except mbustypes.MbusError, e:
	print "reliable send failed:",e.descr
    return 1 # keep command registered

# notifier-timer testfunction
def timer_test( data ):
    print "timer_test"
    m = mbustypes.MMessage( "" ) # empty address means send to ALL entities
    m.payload.append( mbustypes.MCommand( "mbus.python" ) )
    m.payload[-1].args = [ 'python', 'rules', [ 1, 2, [ 1, 2, 3, 4] ], \
		    mbustypes.MSymbol( "somewhat" ) ]
    mbus.send( m )
    return 1 # return true for endless looping,
	     # nothing or false to be removed from scheduler

# one line send
def simple_send( data ):
    # mbus.send( address, commandname, arguments
    mbus.send( "", "simple_send", [ -3.14 ] )
    return 0 # unregister timer callback

# incoming mbus command test
def recv_print( mmsg, data = None ):
    print "received print command, args =", mmsg.payload[0]
    return 1 # keep command registered

def newEntity( maddr ):
    print "New Entity:",
    for key in maddr.keys():
	print "%s:%s"%( key, maddr[key] ),
    print

def lostEntity( maddr ):
    print "Lost Entity:",
    for key in maddr.keys():
	print "%s:%s"%( key, maddr[key] ),
    print

def unknown( mmessage ):
    print "received unknown message:"
    print mmessage.asString()

def transporterror( e ):
    print "Mbus transport error occured:", e.descr
    print "MMessage content:"
    if e.mmsg:
	print e.mmsg.header.srcAdr, e.mmsg.header.destAdr

# create mbus entity
mbus = mbusif.MbusIF( "app:test lang:python" )

# register special callbacks (optional)
mbus.newEntityFunction      = newEntity
mbus.lostEntityFunction     = lostEntity
mbus.unknownMessageFunction = unknown
mbus.errorFunction          = transporterror

mbus.addTimer( 3, simple_send )
mbus.addCallback( "print", recv_print )
mbus.addTimer( 5, sendReliableTest )
mbus.addTimer( 4, timer_test )

try:
    mbus.loop()
#except TypeError, e:
#    print e
except mbustypes.MbusError, e:
    print e.descr
    print e.mmsg
except KeyboardInterrupt:
    pass
